/* sdl-freetype, a text rendering library for SDL.
 * Copyright (C) 2006 luojinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * THANKS FOR ALL CAIRO AUTHORS, MOST OF IDEAS AND IMPLEMENT DTAILS ARE
 * STOLEN FROM CAIRO.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include  "hash.h"
#include  <assert.h>
#include  <stdlib.h>
#include  <stdio.h>
#include  <string.h>

static const int primes[] =
{
    11,
    19,
    37,
    73,
    109,
    163,
    251,
    367,
    557,
    823,
    1237,
    1861,
    2777,
    4177,
    6247,
    9371,
    14057,
    21089,
    31627,
    47431,
    71143,
    106721,
    160073,
    240101,
    360163,
    540217,
    810343,
    1215497,
    1823231,
    2734867,
    4102283,
    6153409,
    9230113,
    13845163,
};
#define PRIMES_SIZE (sizeof(primes) / sizeof(primes[0]))

static int
sdl_freetype_hash_perfect_size (int maxsize)
{
    unsigned int i;

    for (i = 0; i < PRIMES_SIZE; i++)
	if (primes[i] > maxsize)
	    return primes[i];

    return -1;
}

static sdl_freetype_hash_entry_t dead_entry;

sdl_freetype_hash_t *
sdl_freetype_hash_init (sdl_freetype_hash_t * hash, int size,
			const sdl_freetype_hash_ops_t * ops)
{
    int size0;

    if (!ops || !ops->hash || !ops->equal || !ops->destroy)
	goto errquit0;

    size0 = sdl_freetype_hash_perfect_size (size);
    if (size0 < 0)
	goto errquit0;

    hash->entries = malloc (sizeof(sdl_freetype_hash_entry_t *) * size0);
    if (!hash->entries)
	goto errquit1;

    memset (hash->entries, 0, sizeof(sdl_freetype_hash_entry_t *) * size0);
    hash->current_entries = 0;
    hash->size = size0;
    hash->itering = 0;
    hash->ops = *ops;
    return hash;
 errquit1:
    free (hash->entries);
 errquit0:
    return NULL;
}

sdl_freetype_hash_t *
sdl_freetype_hash_create(int size,
			 const sdl_freetype_hash_ops_t * ops)
{
    sdl_freetype_hash_t * hash;

    hash = malloc(sizeof(sdl_freetype_hash_t));
    if (!hash)
	goto errquit0;

    if (!sdl_freetype_hash_init(hash, size, ops))
	goto errquit1;
    return hash;
 errquit1:
    free (hash);
 errquit0:
    return NULL;
}

static sdl_freetype_hash_entry_t **
sdl_freetype_hash_lookup_intl (sdl_freetype_hash_t * hash,
			       sdl_freetype_hash_entry_t * key)
{
    int index, i;
    int first_one = -1;

    index = hash->ops.hash (key) % hash->size;

    for (i = 0; i < hash->size; i++) {
	if (!hash->entries[index]) {
            return &hash->entries[index];
        } else if (hash->entries[index] == &dead_entry) {
            if (first_one < 0)
                first_one = index;
        } else if (hash->ops.equal (key, hash->entries[index])) {
	    return &hash->entries[index];
        }
	if (++index == hash->size)
	    index = 0;
    }

    return first_one < 0 ? NULL : &hash->entries[first_one];
}

sdl_freetype_hash_entry_t *
sdl_freetype_hash_lookup (sdl_freetype_hash_t * hash,
			  sdl_freetype_hash_entry_t * key)
{
    sdl_freetype_hash_entry_t ** entry;

    entry = sdl_freetype_hash_lookup_intl (hash, key);
    if (!entry || *entry == &dead_entry)
	return NULL;

    return *entry;
}

sdl_freetype_hash_entry_t *
sdl_freetype_hash_any_entry (sdl_freetype_hash_t * hash,
			     sdl_freetype_hash_filter_func_t func)
{
    int index, i;

    index = rand() % hash->size;

    if (func) {
	for (i = 0; i < hash->size; i++) {
	    if (hash->entries[index] &&
                hash->entries[index] != &dead_entry &&
                func (hash, hash->entries[index]))
		return hash->entries[index];
	    index += 1;
	    if (index == hash->size)
		index = 0;
	}
    } else {
	for (i = 0; i < hash->size; i++) {
	    if (hash->entries[index] && hash->entries[index] != &dead_entry)
		return hash->entries[index];
	    index += 1;
	    if (index == hash->size)
		index = 0;
	}
    }

    return NULL;
}

static int
sdl_freetype_hash_insert_intl (sdl_freetype_hash_t * hash,
			       sdl_freetype_hash_entry_t * entry,
			       int resize)
{
    sdl_freetype_hash_entry_t ** entry0;

    if (hash->itering)
	return -1;

    entry0 = sdl_freetype_hash_lookup_intl(hash, entry);
    if (!entry0)
	return -1;

    if (*entry0 && *entry0 != &dead_entry)
	return -1;

    hash->current_entries++;
    *entry0 = entry;

    if (resize)
	sdl_freetype_hash_resize(hash);

    return 0;
}

int
sdl_freetype_hash_insert (sdl_freetype_hash_t * hash,
			  sdl_freetype_hash_entry_t * entry)
{
    return sdl_freetype_hash_insert_intl(hash, entry, 1);
}

int
sdl_freetype_hash_remove (sdl_freetype_hash_t * hash,
			  sdl_freetype_hash_entry_t * key)
{
    sdl_freetype_hash_entry_t ** entry, * entry0;

    entry = sdl_freetype_hash_lookup_intl(hash, key);
    if (!entry)
	return -1;

    if (!*entry || *entry == &dead_entry)
	return 0;

    entry0 = *entry;
    *entry = &dead_entry;
    hash->current_entries--;

    hash->ops.destroy(entry0);

    sdl_freetype_hash_resize(hash);

    return 0;
}

int
sdl_freetype_hash_remove_all (sdl_freetype_hash_t * hash)
{
    sdl_freetype_hash_entry_t * entry;
    int i;

    for (i = 0; i < hash->size; i++) {
	if (!hash->entries[i] || hash->entries[i] == &dead_entry)
	    continue;

	entry = hash->entries[i];
	hash->entries[i] = NULL;
	hash->current_entries--;

	hash->ops.destroy (entry);
    }

    return 0;
}

int
sdl_freetype_hash_foreach (sdl_freetype_hash_t * hash,
			   sdl_freetype_hash_callback_func_t func,
			   void * closure)
{
    int i;

    hash->itering++;

    for (i = 0; i < hash->size; i++) {
	if (!hash->entries[i] || hash->entries[i] == &dead_entry)
	    continue;
	func (hash, hash->entries[i], closure);
    }

    hash->itering--;
    if (!hash->itering)
	sdl_freetype_hash_resize (hash);

    return 0;
}

int
sdl_freetype_hash_resize (sdl_freetype_hash_t * hash)
{
    sdl_freetype_hash_t hash0;
    int newsize, i;

    if (hash->itering)
	return 0;

    if (hash->current_entries >= hash->size / 4 &&
	hash->current_entries < hash->size * 3.0 / 4)
	return 0;

    newsize = sdl_freetype_hash_perfect_size(hash->current_entries * 3 / 2);
    if (newsize == hash->size)
	return 0;

    hash0.entries = malloc (sizeof(sdl_freetype_hash_entry_t *) * newsize);
    if (!hash0.entries)
	return -1;

    memset (hash0.entries, 0, sizeof(sdl_freetype_hash_entry_t *) * newsize);
    hash0.current_entries = 0;
    hash0.size = newsize;
    hash0.itering = 0;
    hash0.ops = hash->ops;

    for (i = 0; i < hash->size; i++) {
	if (!hash->entries[i] || hash->entries[i] == &dead_entry)
	    continue;
	sdl_freetype_hash_insert_intl (&hash0, hash->entries[i], 0);
    }

    assert(hash->current_entries == hash0.current_entries);

    free (hash->entries);
    hash->size = newsize;
    hash->entries = hash0.entries;

    return 0;
}

void
sdl_freetype_hash_fini (sdl_freetype_hash_t * hash)
{
    if (!hash)
	return;

    sdl_freetype_hash_remove_all (hash);

    free (hash->entries);
    hash->entries = NULL;
}

void
sdl_freetype_hash_destroy (sdl_freetype_hash_t * hash)
{
    if (!hash)
	return;

    sdl_freetype_hash_fini (hash);
    free (hash);
}
