/* sdl-freetype, a text rendering library for SDL.
 * Copyright (C) 2006 luojinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it either under the terms of the GNU Lesser General Public
 * License version 2.1 as published by the Free Software Foundation
 * (the "LGPL") or, at your option, under the terms of the Mozilla
 * Public License Version 1.1 (the "MPL"). If you do not alter this
 * notice, a recipient may use your version of this file under either
 * the MPL or the LGPL.
 *
 * You should have received a copy of the LGPL along with this library
 * in the file COPYING-LGPL-2.1; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * You should have received a copy of the MPL along with this library
 * in the file COPYING-MPL-1.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY
 * OF ANY KIND, either express or implied. See the LGPL or the MPL for
 * the specific language governing rights and limitations.
 *
 * THANKS FOR ALL CAIRO AUTHORS, MOST OF IDEAS AND IMPLEMENT DTAILS ARE
 * STOLEN FROM CAIRO.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ft2build.h>
#include FT_OUTLINE_H
#include FT_IMAGE_H
#include FT_STROKER_H
#ifdef HAVE_FT_GLYPHSLOT_EMBOLDEN
#include FT_SYNTHESIS_H
#endif
#ifdef HAVE_FT_LIBRARY_SETLCDFILTER
#include FT_LCD_FILTER_H
#endif

#include "cache.h"
#include "sdl-font.h"
#include "sdl-glyph-render.h"
#include "sdl-freetype-utils.h"

#include <math.h>
#include <string.h>

#define DOUBLE_TO_26_6(d) ((FT_F26Dot6)((d) * 64.0))
#define DOUBLE_FROM_26_6(t) ((double)(t) / 64.0)
#define DOUBLE_TO_16_16(d) ((FT_Fixed)((d) * 65536.0))
#define DOUBLE_FROM_16_16(t) ((double)(t) / 65536.0)

#define FIXED_FROM_26_6(t) sdl_freetype_fixed_from_26_6(t)
#define FIXED_FROM_DOUBLE(d) sdl_freetype_fixed_from_double(d)

#ifndef SDL_FREETYPE_GLYPHS_CACHE_SIZE
#define SDL_FREETYPE_GLYPHS_CACHE_SIZE  256
#endif

typedef struct {
    sdl_freetype_cache_entry_t base;
    sdl_freetype_font_t * font;
    sdl_freetype_glyph_info_t info;
} sdl_freetype_glyph_cache_entry_t;

struct _sdl_freetype_font {
    int refcnt;

    sdl_freetype_base_font_t * base_font;
    sdl_freetype_font_face_t * face;
    sdl_freetype_font_options_t options;
    sdl_freetype_cache_t * glyphs;
    sdl_freetype_font_fixed_extents_t metrics;
    sdl_freetype_glyph_render_t * render;
    FT_Stroker stroker;
    FT_Matrix xform;
    double size;
    double stroke_scale;
    int load_flags;
    int load_target;
    int render_mode;
};

static void
sdl_freetype_glyph_entry_destroy (sdl_freetype_glyph_cache_entry_t * entry)
{
    sdl_freetype_font_t * font = entry->font;

    if (entry->info.flags & SDL_FREETYPE_GLYPH_USER_DATA &&
        font->render->glyph_user_data_fini)
        font->render->glyph_user_data_fini (font, font->render,
                                            entry->info.data);

    if (entry->info.flags & SDL_FREETYPE_GLYPH_SURFACE_DATA &&
        entry->info.bitmap.buffer)
        free (entry->info.bitmap.buffer);

    free (entry);
}

static unsigned long
sdl_freetype_glyph_cache_entry_hash (const sdl_freetype_hash_entry_t * entry)
{
    return (unsigned long)entry->key;
}

static int
sdl_freetype_glyph_cache_entry_equal (const sdl_freetype_hash_entry_t * entry0,
                                      const sdl_freetype_hash_entry_t * entry1)
{
    return entry0->key == entry1->key;
}

static void
sdl_freetype_glyph_cache_entry_destroy (sdl_freetype_hash_entry_t * entry)
{
    sdl_freetype_glyph_cache_entry_t * info = (sdl_freetype_glyph_cache_entry_t *) entry;

    sdl_freetype_glyph_entry_destroy (info);
}

static const sdl_freetype_cache_ops_t glyph_cache_ops = {
    sdl_freetype_glyph_cache_entry_hash,
    sdl_freetype_glyph_cache_entry_equal,
    sdl_freetype_glyph_cache_entry_destroy
};

static void
sdl_freetype_font_options_init_default(sdl_freetype_font_options_t * options)
{
    options->antialias = SDL_FT_ANTIALIAS_DEFAULT;
    options->subpixel_order = SDL_FT_SUBPIXEL_ORDER_DEFAULT;
    options->hint = SDL_FT_HINT_STYLE_DEFAULT;
    options->embedded_bitmap = 1;
    options->embolden = 0;
    options->vertical_layout = 0;
    options->autohint = 0;
}

static void
sdl_freetype_font_compute_flags (sdl_freetype_font_t * font)
{
    font->load_flags = FT_LOAD_DEFAULT;
    font->load_target =  FT_LOAD_TARGET_NORMAL;

    if (font->options.antialias == SDL_FT_ANTIALIAS_NONE) {
        font->load_flags |= FT_LOAD_MONOCHROME;
        if (font->options.hint == SDL_FT_HINT_STYLE_NONE)
            font->load_flags |= FT_LOAD_NO_HINTING;
        else
            font->load_target = FT_LOAD_TARGET_MONO;
    } else {
        switch (font->options.hint) {
        case SDL_FT_HINT_STYLE_NONE:
            font->load_flags |= FT_LOAD_NO_HINTING;
            break;
        case SDL_FT_HINT_STYLE_SLIGHT:
            font->load_target = FT_LOAD_TARGET_LIGHT;
            break;
        case SDL_FT_HINT_STYLE_MEDIUM:
            break;
        case SDL_FT_HINT_STYLE_DEFAULT:
        case SDL_FT_HINT_STYLE_FULL:
            if (font->options.antialias == SDL_FT_ANTIALIAS_SUBPIXEL) {
                switch (font->options.subpixel_order) {
                case SDL_FT_SUBPIXEL_ORDER_DEFAULT:
                case SDL_FT_SUBPIXEL_ORDER_RGB:
                case SDL_FT_SUBPIXEL_ORDER_BGR:
                    font->load_target = FT_LOAD_TARGET_LCD;
                    break;
                case SDL_FT_SUBPIXEL_ORDER_VRGB:
                case SDL_FT_SUBPIXEL_ORDER_VBGR:
                    font->load_target = FT_LOAD_TARGET_LCD_V;
                    break;
                }
            }
            break;
        }

        if (font->options.autohint)
            font->load_flags |= FT_LOAD_FORCE_AUTOHINT;
        if (!font->options.embedded_bitmap)
            font->load_flags |= FT_LOAD_NO_BITMAP;
    }

    switch (font->options.antialias) {
    case SDL_FT_ANTIALIAS_NONE:
        font->render_mode = FT_RENDER_MODE_MONO;
        break;
    case SDL_FT_ANTIALIAS_SUBPIXEL:
        if (font->options.subpixel_order == SDL_FT_SUBPIXEL_ORDER_VRGB ||
            font->options.subpixel_order == SDL_FT_SUBPIXEL_ORDER_VBGR)
            font->render_mode = FT_RENDER_MODE_LCD_V;
        else
            font->render_mode = FT_RENDER_MODE_LCD;
        break;
    case SDL_FT_ANTIALIAS_DEFAULT:
    case SDL_FT_ANTIALIAS_GRAY:
        font->render_mode = FT_RENDER_MODE_NORMAL;
        break;
    }

}

sdl_freetype_font_t *
sdl_freetype_font_create(const char * file, double size)
{
    return sdl_freetype_font_create_full (file, size, NULL, 0);
}

static void
sdl_freetype_font_setup_options (sdl_freetype_font_t * font,
                                 const sdl_freetype_font_desc_t * desc,
                                 unsigned int mask)
{
#ifndef HAVE_FT_GLYPHSLOT_EMBOLDEN
    mask &= ~SDL_FONT_EMBOLDEN_MASK;
#endif

#ifndef HAVE_FT_LIBRARY_SETLCDFILTER
    if (mask & SDL_FONT_ANTIALIAS_MASK &&
        desc->antialias == SDL_FT_ANTIALIAS_SUBPIXEL)
        mask &= ~SDL_FONT_ANTIALIAS_MASK;
    mask &= ~SDL_FONT_SUBPIXEL_ORDER_MASK;
#endif

    sdl_freetype_font_options_init_default (&font->options);
    if (mask & SDL_FONT_ANTIALIAS_MASK)
        font->options.antialias = desc->antialias;
    if (mask & SDL_FONT_SUBPIXEL_ORDER_MASK)
        font->options.subpixel_order = desc->subpixel_order;
    if (mask & SDL_FONT_EMBEDDEDBITMAP_MASK)
        font->options.embolden = desc->embedded_bitmap;
    if (mask & SDL_FONT_EMBOLDEN_MASK)
        font->options.embolden = desc->embolden;
    if (mask & SDL_FONT_HINT_STYLE_MASK)
        font->options.hint = desc->hint;
    if (mask & SDL_FONT_VERTICAL_LAYOUT_MASK)
        font->options.vertical_layout = desc->vertical_layout;
    if (mask & SDL_FONT_AUTOHINT_MASK)
        font->options.autohint = desc->autohint;
    if (mask & SDL_FONT_TRANSFORM_MASK) {
        font->xform = desc->xform;
    } else {
        font->xform.xx = 1 << 16;
        font->xform.yx = 0;
        font->xform.xy = 0;
        font->xform.yy = 1 << 16;
    }

    if (mask & SDL_FONT_GLYPH_RENDER_MASK)
        font->render = desc->render;
    else
        font->render = &sdl_freetype_default_glyph_render;

    if (mask & SDL_FONT_STROKE_SCALE_MASK)
        font->stroke_scale = desc->stroke_scale;
    else
        font->stroke_scale = 1.0;
}

sdl_freetype_font_t *
sdl_freetype_font_create_full (const char * file, double size,
                               const sdl_freetype_font_desc_t * desc,
                               unsigned int mask)
{
    sdl_freetype_base_font_t * base_font;
    sdl_freetype_font_t * font;
    int index = 0;
    FT_Face face;
    double scale;
    sdl_freetype_font_fixed_extents_t * metrics;

    if (mask & SDL_FONT_INDEX_MASK)
        index = desc->index;
    if (mask & SDL_FONT_FROM_MEMORY_MASK)
        base_font = sdl_freetype_base_font_create_from_memory (desc->data,
                                                               desc->size,
                                                               index);
    else
        base_font = sdl_freetype_base_font_create (file, index);
    if (!base_font)
        goto errquit0;
    font = malloc (sizeof(sdl_freetype_font_t));
    if (!font)
        goto errquit1;
    font->glyphs = sdl_freetype_cache_create (SDL_FREETYPE_GLYPHS_CACHE_SIZE,
                                              &glyph_cache_ops);
    if (!font->glyphs)
        goto errquit2;
    font->base_font = base_font;
    font->size = size;
    font->refcnt = 1;

    sdl_freetype_font_setup_options (font, desc, mask);
    font->face = sdl_freetype_font_face_create (font->base_font,
                                                &font->options);
    if (!font->face)
        goto errquit3;
    face = sdl_freetype_font_lock_face (font);
    if (!face)
        goto errquit4;

    /* compute load flags & load target */
    sdl_freetype_font_compute_flags (font);

    /* compute font metrics */
    metrics = &font->metrics;

    scale = face->units_per_EM;
    metrics->ascent  = FIXED_FROM_DOUBLE(face->ascender * font->size / scale);
    metrics->descent = FIXED_FROM_DOUBLE(-face->descender * font->size / scale);
    metrics->height  = FIXED_FROM_DOUBLE(face->height * font->size / scale);
    if (!font->options.vertical_layout) {
        metrics->max_x_advance =
            FIXED_FROM_DOUBLE(face->max_advance_width * font->size / scale);
        metrics->max_y_advance = 0;
    } else {
        metrics->max_x_advance = 0;
        metrics->max_y_advance =
            FIXED_FROM_DOUBLE(face->max_advance_height * font->size / scale);
    }

    if (mask & SDL_FONT_STROKE_MASK && desc->stroke) {
        FT_Error error;
        FT_Library library =
            sdl_freetype_base_font_get_freetype_library (font->base_font);
        if (!library)
            goto errquit5;
#if (FREETYPE_MAJOR > 2) || ((FREETYPE_MAJOR == 2) && (FREETYPE_MINOR > 1))
        error = FT_Stroker_New (library, &font->stroker );
#else
        error = FT_Stroker_New (face->memory, &font->stroke);
#endif
        if (error)
            goto errquit5;
        FT_Stroker_Set (font->stroker,
                        64 * font->stroke_scale,
                        FT_STROKER_LINECAP_ROUND,
                        FT_STROKER_LINEJOIN_ROUND,
                        0);
    } else {
        font->stroker = NULL;
    }
    sdl_freetype_font_unlock_face (font);
    return font;
 errquit5:
    sdl_freetype_font_unlock_face (font);
 errquit4:
    sdl_freetype_font_face_destroy (font->face);
 errquit3:
    sdl_freetype_cache_destroy (font->glyphs);
 errquit2:
    free (font);
 errquit1:
    sdl_freetype_base_font_destroy (base_font);
 errquit0:
    return NULL;
}

sdl_freetype_font_t *
sdl_freetype_font_reference (sdl_freetype_font_t * font)
{
    if (!font)
        return font;

    font->refcnt++;

    return font;
}

FT_Face
sdl_freetype_font_lock_face (sdl_freetype_font_t * font)
{
    FT_Face face;

    if (!font)
        return NULL;

    face = sdl_freetype_base_font_lock_face (font->base_font);
    if (!face)
        return NULL;

    sdl_freetype_base_font_select_size (font->base_font,
                                        font->size,
                                        &font->xform);

    return face;
}

void
sdl_freetype_font_unlock_face (sdl_freetype_font_t * font)
{
    if (!font)
        return;
    sdl_freetype_base_font_unlock_face (font->base_font);
}

FT_UInt
sdl_freetype_font_glyph_index (sdl_freetype_font_t * font,
                               FT_Int32 ucs4)
{
    FT_Face face;
    FT_UInt index;

    face = sdl_freetype_font_lock_face (font);
    if (!face)
        return 0;
    index = FT_Get_Char_Index (face, ucs4);
    sdl_freetype_font_unlock_face (font);
    return index;
}

#ifndef HAVE_FT_GLYPHSLOT_EMBOLDEN
#define FT_GlyphSlot_Embolden(g) ((void)0)
#endif

#ifndef HAVE_FT_LIBRARY_SETLCDFILTER
#define FT_Library_SetLcdFilter(l, f) ((void)0)
#endif

static int
ft_target_bitmap_init (FT_Bitmap *     target,
                       FT_Bitmap *     bitmap,
                       FT_Render_Mode  mode)
{
    unsigned         width, height, pitch;

    width  = bitmap->width;
    height = bitmap->rows;
    pitch  = (width + 3) & ~3;

    switch (bitmap->pixel_mode)
    {
    case FT_PIXEL_MODE_MONO:
        if (mode == FT_RENDER_MODE_MONO ) {
            pitch = (((width + 31) & ~31) >> 3);
            break;
        }
        /* fall-through */
    case FT_PIXEL_MODE_GRAY:
        if (mode == FT_RENDER_MODE_LCD || mode == FT_RENDER_MODE_LCD_V)
            pitch = width * 4;
        break;
    case FT_PIXEL_MODE_LCD:
        if (mode != FT_RENDER_MODE_LCD)
            return -1;

        width /= 3;
        pitch = width * 4;
        break;
    case FT_PIXEL_MODE_LCD_V:
        if (mode != FT_RENDER_MODE_LCD_V)
            return -1;
        height /= 3;
        pitch  = width * 4;
        break;
    default:
        return -1;
    }

    target->width  = width;
    target->rows   = height;
    target->pitch  = pitch;
    target->buffer = calloc (pitch * height, 1);

    if (!target->buffer)
        return -1;
    return 0;
}

#if 0
static void
dump_mono_bitmap(FT_Bitmap * bm, int left, int top)
{
    int x, y, bit;
    int rows   = bm->rows;
    int width  = bm->width;
    int pitch  = bm->pitch;
    unsigned char * buffer = bm->buffer;
    unsigned char * p = buffer;

    printf("rows = %d, width = %d, pitch = %d, bufer = %p\n", rows, width, pitch, buffer );
    printf("left = %d top = %d\n", left, top );

    for (y = 0; y < rows; y++) {
        for (x = 0; x < pitch; x++) {
            for (bit = 0; bit < 8; bit++) {
                int b = * p;
                int v = b << bit;
                if (v & 0x80 )
                    printf(" *" );
                else
                    printf("  ");
            }
            p++;
        }
        printf("\n");
    }
}

static void
dump_gray_bitmap(FT_Bitmap * bm, int left, int top)
{
    int x, y;
    int rows   = bm->rows;
    int width  = bm->width;
    int pitch  = bm->pitch;
    unsigned char * buffer = bm->buffer;
    unsigned char * p = buffer;

    printf("rows = %d, width = %d, pitch = %d, bufer = %p\n", rows, width, pitch, buffer );
    printf("left = %d top = %d\n", left, top );

    for (y = 0; y < rows; y++) {
        for (x = 0; x < width; x++)
            printf(" %03d", p[width]);
        p += pitch;
        printf("\n");
    }
}

static void
dump_bitmap(FT_Bitmap * bm, int left, int top)
{
    if (bm->pixel_mode == FT_PIXEL_MODE_MONO)
        dump_mono_bitmap(bm, left, top);
    else
        dump_gray_bitmap(bm, left, top);
}
#endif

static void
ft_render_target_bitmap (FT_Bitmap *     target,
                         FT_Bitmap *     bitmap,
                         FT_Render_Mode  mode,
                         int             bgr)
{
    unsigned char *  srcline   = bitmap->buffer;
    unsigned char *  dstline   = target->buffer;
    int              src_pitch = bitmap->pitch;
    int              width     = target->width;
    int              height    = target->rows;
    int              pitch     = target->pitch;
    int              subpixel;
    int              h;

    if (mode == FT_RENDER_MODE_LCD || mode == FT_RENDER_MODE_LCD_V)
        subpixel = 1;
    else
        subpixel = 0;

    if (src_pitch < 0)
        srcline -= src_pitch * (bitmap->rows - 1);

    target->pixel_mode = bitmap->pixel_mode;
    switch (bitmap->pixel_mode)
    {
    case FT_PIXEL_MODE_MONO:
        if (subpixel) {
            /* convert mono to ARGB32 values */
            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch) {
                int  x;

                for (x = 0; x < width; x++) {
                    if (srcline[(x >> 3)] & (0x80 >> (x & 7)))
                        ((unsigned int*)dstline)[x] = 0xffffffffU;
                }
            }
            target->pixel_mode = FT_PIXEL_MODE_LCD;
        } else if (mode == FT_RENDER_MODE_NORMAL) {
            /* convert mono to 8-bit gray */

            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch) {
                int  x;

                for (x = 0; x < width; x++) {
                    if (srcline[(x >> 3)] & (0x80 >> (x & 7)))
                        dstline[x] = 0xff;
                }
            }
            target->pixel_mode = FT_PIXEL_MODE_GRAY;
        } else {
            /* copy mono to mono */
            int  bytes = (width + 7) >> 3;

            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch)
                memcpy(dstline, srcline, bytes);
        }
        break;
    case FT_PIXEL_MODE_GRAY:
        if (subpixel) {
            /* convert gray to ARGB32 values */
            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch ) {
                int            x;
                unsigned int*  dst = (unsigned int *)dstline;

                for ( x = 0; x < width; x++ ) {
                    unsigned int  pix = srcline[x];

                    pix |= (pix << 8);
                    pix |= (pix << 16);

                    dst[x] = pix;
                }
            }
            target->pixel_mode = FT_PIXEL_MODE_LCD;
        } else {
            /* copy gray into gray */
            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch)
                memcpy(dstline, srcline, width);
        }
        break;
    case FT_PIXEL_MODE_LCD:
        if (!bgr) {
            /* convert horizontal RGB into ARGB32 */
            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch ) {
                int             x;
                unsigned char * src = srcline;
                unsigned int *  dst = (unsigned int *)dstline;

                for (x = 0; x < width; x++, src += 3) {
                    unsigned int  pix;

                    pix =
                        ((unsigned int)src[0] << 16) |
                        ((unsigned int)src[1] <<  8) |
                        ((unsigned int)src[2]      ) |
                        ((unsigned int)src[1] << 24);

                    dst[x] = pix;
                }
            }
        } else {
            /* convert horizontal BGR into ARGB32 */
            for (h = height; h > 0; h--, srcline += src_pitch, dstline += pitch ) {
                int             x;
                unsigned char * src = srcline;
                unsigned int *  dst = (unsigned int *)dstline;

                for (x = 0; x < width; x++, src += 3) {
                    unsigned int  pix;

                    pix =
                        ((unsigned int)src[2] << 16) |
                        ((unsigned int)src[1] <<  8) |
                        ((unsigned int)src[0]      ) |
                        ((unsigned int)src[1] << 24);

                    dst[x] = pix;
                }
            }
        }
        break;
    default:  /* FT_PIXEL_MODE_LCD_V */
        /* convert vertical RGB into ARGB32 */
        if (!bgr) {
            for (h = height; h > 0; h--, srcline += 3 * src_pitch, dstline += pitch) {
                int            x;
                unsigned char * src = srcline;
                unsigned int *  dst = (unsigned int*)dstline;

                for (x = 0; x < width; x++, src += 1) {
                    unsigned int  pix;
#if 1
                    pix =
                        ((unsigned int)src[0]           << 16) |
                        ((unsigned int)src[src_pitch]   <<  8) |
                        ((unsigned int)src[src_pitch * 2]    ) |
                        0xFF000000 ;
#else
                    pix =
                        ((unsigned int)src[0]           << 16) |
                        ((unsigned int)src[src_pitch]   <<  8) |
                        ((unsigned int)src[src_pitch * 2]    ) |
                        ((unsigned int)src[src_pitch]   << 24);
#endif
                    dst[x] = pix;
                }
            }
        } else {
            for (h = height; h > 0; h--, srcline += 3 * src_pitch, dstline += pitch) {
                int             x;
                unsigned char * src = srcline;
                unsigned int *  dst = (unsigned int*)dstline;

                for (x = 0; x < width; x++, src += 1) {
                    unsigned int  pix;

                    pix =
                        ((unsigned int)src[src_pitch * 2] << 16) |
                        ((unsigned int)src[src_pitch]     <<  8) |
                        ((unsigned int)src[0]                  ) |
                        ((unsigned int)src[src_pitch]     << 24);

                    dst[x] = pix;
                }
            }
        }
    }
}

static int
sdl_freetype_font_glyph_info_render_glyph (sdl_freetype_font_t * font,
                                           sdl_freetype_glyph_info_t * glyph_info,
                                           FT_Face face, FT_UInt index,
                                           FT_Bitmap * bitmap,
                                           int bitmap_left, int bitmap_top)
{
    FT_Bitmap target;
    int bgr;

    if (ft_target_bitmap_init (&target, bitmap, font->render_mode) < 0)
        return -1;

    if (font->options.antialias == SDL_FT_ANTIALIAS_SUBPIXEL &&
        (font->options.subpixel_order == SDL_FT_SUBPIXEL_ORDER_BGR ||
         font->options.subpixel_order == SDL_FT_SUBPIXEL_ORDER_VBGR))
        bgr = 1;
    else
        bgr = 0;
    ft_render_target_bitmap (&target, bitmap, font->render_mode, bgr);
    glyph_info->flags |=
        SDL_FREETYPE_GLYPH_SURFACE | SDL_FREETYPE_GLYPH_SURFACE_DATA;
    glyph_info->bitmap = target;
    glyph_info->x_offset = bitmap_left;
    glyph_info->y_offset = -bitmap_top;

    return 0;
}

static int
sdl_freetype_font_glyph_entry_info_load (sdl_freetype_font_t * font,
                                         sdl_freetype_glyph_cache_entry_t * entry,
                                         FT_UInt index,
                                         sdl_freetype_glyph_flags_t what)
{
    sdl_freetype_glyph_info_t * glyph_info;
    int load_flags;
    FT_Face face;
    FT_GlyphSlot glyph;
    FT_Error error;

    face = sdl_freetype_font_lock_face (font);
    if (!face)
        goto errquit0;

    load_flags = font->load_flags | FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH;
    error = FT_Load_Glyph (face, index, load_flags);
    if (error) {
        printf ("<%s> load glyph error, font %p face %p %x %x\n", __FUNCTION__,
                font, font->face, index, load_flags);
        goto errquit1;
    }

    glyph = face->glyph;

    if (font->options.embolden)
        FT_GlyphSlot_Embolden (glyph);

    if (font->options.vertical_layout) {
        FT_Vector vector;

        vector.x = glyph->metrics.vertBearingX - glyph->metrics.horiBearingX;
        vector.y = -glyph->metrics.vertBearingY - glyph->metrics.horiBearingY;

        if (glyph->format == FT_GLYPH_FORMAT_OUTLINE) {
            FT_Matrix matrix;

            matrix.xx = font->xform.xx;
            matrix.xy = -font->xform.xy;
            matrix.yx = -font->xform.yx;
            matrix.yy = font->xform.yy;
            FT_Vector_Transform (&vector, &matrix);
            FT_Outline_Translate(&glyph->outline, vector.x, vector.y);
        } else if (glyph->format == FT_GLYPH_FORMAT_BITMAP) {
            glyph->bitmap_left += vector.x / 64;
            glyph->bitmap_top  += vector.y / 64;
        }
    }

    glyph_info = &entry->info;
    if (what & SDL_FREETYPE_GLYPH_METRICS) {
        FT_Glyph_Metrics * metrics = &glyph->metrics;

        glyph_info->flags |= SDL_FREETYPE_GLYPH_METRICS;

        glyph_info->metrics.width  = FIXED_FROM_26_6 (metrics->width);
        glyph_info->metrics.height = FIXED_FROM_26_6 (metrics->height);

        if (!font->options.vertical_layout) {
            glyph_info->metrics.x_bearing = FIXED_FROM_26_6 (metrics->horiBearingX);
            glyph_info->metrics.y_bearing = FIXED_FROM_26_6 (-metrics->horiBearingY);

            glyph_info->metrics.x_advance = FIXED_FROM_26_6 (metrics->horiAdvance);
            glyph_info->metrics.y_advance = 0;
        } else {
            glyph_info->metrics.x_bearing = FIXED_FROM_26_6 (metrics->vertBearingX);
            glyph_info->metrics.y_bearing = FIXED_FROM_26_6 (metrics->vertBearingY);

            glyph_info->metrics.x_advance = 0;
            glyph_info->metrics.y_advance = FIXED_FROM_26_6 (metrics->vertAdvance);
        }

        if (glyph_info->metrics.width && glyph_info->metrics.height) {
            glyph_info->bbox.x1 = glyph_info->metrics.x_bearing;
            glyph_info->bbox.y1 = glyph_info->metrics.y_bearing;
            glyph_info->bbox.x2 =
                glyph_info->metrics.x_bearing + glyph_info->metrics.width;
            glyph_info->bbox.y2 =
                glyph_info->metrics.y_bearing + glyph_info->metrics.height;
        } else {
            glyph_info->bbox.x1 = glyph_info->metrics.x_bearing;
            glyph_info->bbox.y1 = glyph_info->metrics.y_bearing;
            glyph_info->bbox.x2 =
                glyph_info->metrics.x_bearing + glyph_info->metrics.x_advance;
            glyph_info->bbox.y2 =
                glyph_info->metrics.y_bearing + glyph_info->metrics.y_advance;
        }

        if (glyph_info->metrics.width == 0 || glyph_info->metrics.height == 0)
            glyph_info->flags |= SDL_FREETYPE_GLYPH_EMPTY;
    }

    if (what & SDL_FREETYPE_GLYPH_SURFACE) {
        if (!(glyph_info->flags & SDL_FREETYPE_GLYPH_EMPTY)) {
            int ret;

            if (font->stroker) {
                if (glyph->format == FT_GLYPH_FORMAT_BITMAP)
                    goto errquit1;
                else {
                    FT_Glyph olglyph;
                    FT_BitmapGlyph bpglyph;

                    error = FT_Get_Glyph (glyph, &olglyph);
                    if (error)
                        goto errquit1;
                    error = FT_Glyph_Stroke (&olglyph,
                                             font->stroker, 1);
                    if (error) {
                        FT_Done_Glyph (olglyph);
                        goto errquit1;
                    }
                    error = FT_Glyph_To_Bitmap (&olglyph, font->render_mode,
                                                NULL, 1);
                    if (error) {
                        FT_Done_Glyph (olglyph);
                        goto errquit1;
                    }
                    bpglyph = (FT_BitmapGlyph)olglyph;
                    ret =
                        sdl_freetype_font_glyph_info_render_glyph (font,
                                                                   glyph_info,
                                                                   face, index,
                                                                   &bpglyph->bitmap,
                                                                   bpglyph->left,
                                                                   bpglyph->top);
                    FT_Done_Glyph (olglyph);
                    if (ret < 0)
                        goto errquit1;
                }
            } else {
                if (font->render_mode == FT_RENDER_MODE_LCD ||
                    font->render_mode == FT_RENDER_MODE_LCD) {
                    FT_Library_SetLcdFilter(glyph->library, FT_LCD_FILTER_DEFAULT);
                    error = FT_Render_Glyph (glyph, font->render_mode);
                    FT_Library_SetLcdFilter(glyph->library, FT_LCD_FILTER_NONE);
                } else {
                    error = FT_Render_Glyph (glyph, font->render_mode);
                }
                if (error)
                    goto errquit1;
                ret =
                    sdl_freetype_font_glyph_info_render_glyph (font,
                                                               glyph_info,
                                                               face, index,
                                                               &glyph->bitmap,
                                                               glyph->bitmap_left,
                                                               glyph->bitmap_top);
                if (ret < 0)
                    goto errquit1;
            }
        }
    }
    sdl_freetype_font_unlock_face (font);
    return 0;
 errquit1:
    fprintf (stderr, "<%s> load glyph error, font %p face %p %x %x\n", __FUNCTION__,
             font, font->face, index, load_flags);
    sdl_freetype_font_unlock_face (font);
 errquit0:
    return -1;
}

static int
sdl_freetype_font_glyph_entry_init_user_data (sdl_freetype_font_t * font,
                                              sdl_freetype_glyph_cache_entry_t * entry,
                                              FT_UInt index)
{
    if (entry->info.flags & SDL_FREETYPE_GLYPH_USER_DATA)
        return 0;

    if (font->render->glyph_user_data_init &&
        font->render->glyph_user_data_init (font, font->render,
                                            &entry->info, index))
        return -1;

    entry->info.flags |= SDL_FREETYPE_GLYPH_USER_DATA;
    return 0;
}

static sdl_freetype_glyph_cache_entry_t *
sdl_freetype_font_glyph_entry_lookup (sdl_freetype_font_t * font, FT_UInt index,
                                      sdl_freetype_glyph_flags_t what)
{
    sdl_freetype_glyph_cache_entry_t key, * entry;

    key.base.base.key = (void *)((unsigned long)index);
    key.base.size = 1;

    if (what & SDL_FREETYPE_GLYPH_SURFACE_DATA)
        what |= SDL_FREETYPE_GLYPH_SURFACE;

    entry = (sdl_freetype_glyph_cache_entry_t *)
        sdl_freetype_cache_lookup (font->glyphs, &key.base);
    if (entry) {
        what &= ~SDL_FREETYPE_GLYPH_METRICS;
        /* load glyph surface */
        if (!(entry->info.flags & SDL_FREETYPE_GLYPH_EMPTY) &&
            !(entry->info.flags & SDL_FREETYPE_GLYPH_SURFACE_DATA) &&
            what & SDL_FREETYPE_GLYPH_SURFACE) {
            sdl_freetype_font_glyph_entry_info_load (font, entry, index, what);
        }
        /* init glyph user data */
        if (what & SDL_FREETYPE_GLYPH_USER_DATA)
            sdl_freetype_font_glyph_entry_init_user_data (font, entry, index);
        return entry;
    }

    entry = malloc (sizeof(sdl_freetype_glyph_cache_entry_t));
    if (!entry)
        return NULL;

    memset (entry, 0, sizeof(sdl_freetype_glyph_cache_entry_t));
    entry->base.base.key = (void *)((unsigned long)index);
    entry->base.size = 1;
    entry->font = font;

    what |= SDL_FREETYPE_GLYPH_METRICS;
    if (sdl_freetype_font_glyph_entry_info_load (font, entry, index, what) < 0)
        goto errquit0;

    if (sdl_freetype_cache_insert (font->glyphs, &entry->base) < 0)
        goto errquit0;

    if (what & SDL_FREETYPE_GLYPH_USER_DATA)
        sdl_freetype_font_glyph_entry_init_user_data (font, entry, index);

    return entry;
 errquit0:
    sdl_freetype_glyph_entry_destroy (entry);
    return NULL;
}

sdl_freetype_glyph_info_t *
sdl_freetype_font_glyph_info_lookup (sdl_freetype_font_t * font, FT_UInt index,
                                     sdl_freetype_glyph_flags_t what)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index, what);
    if (!entry)
        return NULL;
    return &entry->info;
}

static int
sdl_freetype_font_glyph_metrics_and_bbox (sdl_freetype_font_t * font, FT_UInt index,
                                          sdl_freetype_text_fixed_extents_t * metrics,
                                          sdl_freetype_bbox_t * bbox)
{
    sdl_freetype_glyph_cache_entry_t * cache;

    cache = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_METRICS);
    if (!cache)
        return -1;
    *metrics = cache->info.metrics;
    *bbox = cache->info.bbox;
    return 0;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_fixed_metrics (sdl_freetype_font_t * font, FT_UInt index,
                                       sdl_freetype_text_fixed_extents_t * metrics)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_METRICS);
    if (!entry)
        return SDL_FREETYPE_ERR_NOT_FOUND;
    *metrics = entry->info.metrics;
    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_metrics (sdl_freetype_font_t * font, FT_UInt index,
                                 sdl_freetype_text_extents_t * metrics)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_METRICS);
    if (!entry)
        return SDL_FREETYPE_ERR_NOT_FOUND;
    sdl_freetype_text_extents_from_fixed(metrics, &entry->info.metrics);
    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_surface (sdl_freetype_font_t * font, FT_UInt index,
                                 FT_Bitmap ** surface, int * x_offset,
                                 int * y_offset, void ** data)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_SURFACE |
                                                  SDL_FREETYPE_GLYPH_SURFACE_DATA);
    if (!entry)
        return SDL_FREETYPE_ERR_NOT_FOUND;
    if (!(entry->info.flags & SDL_FREETYPE_GLYPH_SURFACE))
        return SDL_FREETYPE_ERR_OUT_OF_RESOURCES;

    if (surface) {
        if (entry->info.flags & SDL_FREETYPE_GLYPH_EMPTY)
            *surface = NULL;
        else
            *surface = &entry->info.bitmap;
    }
    if (x_offset)
        *x_offset = entry->info.x_offset;
    if (y_offset)
        *y_offset = entry->info.y_offset;
    if (data)
        *data = entry->info.data;

    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_free_surface_data (sdl_freetype_font_t * font, FT_UInt index,
                                           sdl_freetype_glyph_info_t * info)
{
    if (!info) {
        sdl_freetype_glyph_cache_entry_t * entry;

        entry = sdl_freetype_font_glyph_entry_lookup (font, index, SDL_FREETYPE_GLYPH_NONE);
        if (entry)
            info = &entry->info;
    }

    if (info && info->flags & SDL_FREETYPE_GLYPH_SURFACE_DATA) {
        info->flags &= ~SDL_FREETYPE_GLYPH_SURFACE_DATA;
        if (info->bitmap.buffer)
            free (info->bitmap.buffer);
        info->bitmap.buffer = NULL;
    }

    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_get_user_data (sdl_freetype_font_t * font,
                                       FT_UInt index, void ** data)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_USER_DATA);
    if (!entry)
        return SDL_FREETYPE_ERR_NOT_FOUND;
    if (!(entry->info.flags & SDL_FREETYPE_GLYPH_USER_DATA))
        return SDL_FREETYPE_ERR_UNKNOWN;

    *data = entry->info.data;
    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyph_clear_user_data (sdl_freetype_font_t * font,
                                         FT_UInt index)
{
    sdl_freetype_glyph_cache_entry_t * entry;

    entry = sdl_freetype_font_glyph_entry_lookup (font, index,
                                                  SDL_FREETYPE_GLYPH_NONE);
    if (!entry)
        return SDL_FREETYPE_ERR_NOT_FOUND;
    if (!(entry->info.flags & SDL_FREETYPE_GLYPH_USER_DATA))
        return SDL_FREETYPE_ERR_OK;

    if (font->render->glyph_user_data_fini)
        font->render->glyph_user_data_fini (font, font->render,
                                            entry->info.data);

    entry->info.data = NULL;
    entry->info.flags &= ~SDL_FREETYPE_GLYPH_USER_DATA;
    return SDL_FREETYPE_ERR_OK;
}

static void
sdl_freetype_font_glyph_clear_user_data_cb (sdl_freetype_cache_t * cache,
                                            sdl_freetype_cache_entry_t * entry,
                                            void * closure)
{
    sdl_freetype_glyph_cache_entry_t * glyph_entry =
        (sdl_freetype_glyph_cache_entry_t *) entry;
    sdl_freetype_font_t * font = (sdl_freetype_font_t *) closure;

    if (!(glyph_entry->info.flags & SDL_FREETYPE_GLYPH_USER_DATA))
        return;

    if (font->render->glyph_user_data_fini)
        font->render->glyph_user_data_fini (font, font->render,
                                            glyph_entry->info.data);

    glyph_entry->info.flags &= ~SDL_FREETYPE_GLYPH_USER_DATA;
}

static int
sdl_freetype_font_glyph_clear_all_user_data (sdl_freetype_font_t * font)
{

    sdl_freetype_cache_foreach (font->glyphs,
                                sdl_freetype_font_glyph_clear_user_data_cb,
                                font);
    return 0;
}

static sdl_freetype_glyph_render_t sdl_freetype_dummy_glyph_render = {
    SDL_FT_RENDER_CAPS_NONE, NULL, NULL, NULL, NULL
};

sdl_freetype_error_t
sdl_freetype_font_set_render (sdl_freetype_font_t * font,
                              sdl_freetype_glyph_render_t * render)
{
    if (font->render == render)
        return SDL_FREETYPE_ERR_OK;
    if (font->render == &sdl_freetype_dummy_glyph_render && !render)
        return SDL_FREETYPE_ERR_OK;

    sdl_freetype_font_glyph_clear_all_user_data (font);
    if (font->render->destroy)
        font->render->destroy (font, render);

    font->render = render ? render : &sdl_freetype_dummy_glyph_render;

    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_glyph_render_t *
sdl_freetype_font_get_render (sdl_freetype_font_t * font)
{
    if (font->render == &sdl_freetype_dummy_glyph_render)
        return NULL;
    return font->render;
}

void
sdl_freetype_font_glyph_freeze (sdl_freetype_font_t * font)
{
    sdl_freetype_cache_freeze (font->glyphs);
}

void
sdl_freetype_font_glyph_thaw (sdl_freetype_font_t * font)
{
    sdl_freetype_cache_thaw (font->glyphs);
}

sdl_freetype_error_t
sdl_freetype_font_fixed_extents (sdl_freetype_font_t * font,
                                 sdl_freetype_font_fixed_extents_t * metrics)
{
    *metrics = font->metrics;
    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_extents (sdl_freetype_font_t * font,
                           sdl_freetype_font_extents_t * metrics)
{
    metrics->ascent = sdl_freetype_fixed_to_double(font->metrics.ascent);
    metrics->descent = sdl_freetype_fixed_to_double(font->metrics.descent);
    metrics->height = sdl_freetype_fixed_to_double(font->metrics.height);
    metrics->max_x_advance = sdl_freetype_fixed_to_double(font->metrics.max_x_advance);
    metrics->max_y_advance = sdl_freetype_fixed_to_double(font->metrics.max_y_advance);

    return SDL_FREETYPE_ERR_OK;
}

void
sdl_freetype_font_destroy (sdl_freetype_font_t * font)
{
    if (!font)
        return;

    font->refcnt--;
    if (!font->refcnt) {
        sdl_freetype_cache_empty (font->glyphs);
        if (font->render->destroy)
            font->render->destroy (font, font->render);
        sdl_freetype_cache_destroy (font->glyphs);
        if (font->stroker)
            FT_Stroker_Done (font->stroker);
        sdl_freetype_font_face_destroy (font->face);
        sdl_freetype_base_font_destroy (font->base_font);
        free (font);
    }
}

#define LOCAL_BUFSZ (64)

static sdl_freetype_error_t
sdl_freetype_font_show_glyphs_intl (sdl_freetype_font_t * font,
                                    void * dst,
                                    unsigned char r,
                                    unsigned char g,
                                    unsigned char b,
                                    unsigned char a,
                                    int x, int y,
                                    const sdl_freetype_glyph_t * glyphs,
                                    int num)
{
    sdl_freetype_glyph_cache_entry_t * entry;
    int i;

    if (!font->render->glyph_render)
        return SDL_FREETYPE_ERR_OK;

    sdl_freetype_font_glyph_freeze (font);

    for (i = 0; i < num; i++) {
        entry = sdl_freetype_font_glyph_entry_lookup (font, glyphs[i].index,
                                                      SDL_FREETYPE_GLYPH_METRICS |
                                                      SDL_FREETYPE_GLYPH_SURFACE);
        if (!entry || !(entry->info.flags & SDL_FREETYPE_GLYPH_SURFACE))
            continue;

        font->render->glyph_render  (font, dst,
                                     r, g, b, a,
                                     x + sdl_freetype_fixed_round(glyphs[i].x),
                                     y + sdl_freetype_fixed_round(glyphs[i].y),
                                     glyphs[i].index,
                                     &entry->info.bitmap,
                                     entry->info.x_offset,
                                     entry->info.y_offset);
    }

    sdl_freetype_font_glyph_thaw (font);

    return 0;
}

sdl_freetype_error_t
sdl_freetype_font_show_glyphs (sdl_freetype_font_t * font,
                               void * dst,
                               unsigned char r,
                               unsigned char g,
                               unsigned char b,
                               unsigned char a,
                               int x, int y,
                               const sdl_freetype_glyph_t * glyphs,
                               int num_glyphs)
{
    sdl_freetype_error_t ret;

    if (!a)
        return SDL_FREETYPE_ERR_OK;

    ret = 0;
    if (font->render->glyphs_render) {
        ret = font->render->glyphs_render (font, dst, r, g, b, a,
                                           x, y, glyphs, num_glyphs);
    } else {
        sdl_freetype_font_glyph_freeze (font);

        ret = sdl_freetype_font_show_glyphs_intl (font, dst, r, g, b, a,
                                                  x, y, glyphs, num_glyphs);

        sdl_freetype_font_glyph_thaw (font);
    }

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_glyphs_fixed_extents (sdl_freetype_font_t * font,
                                        sdl_freetype_text_fixed_extents_t * extents,
                                        const sdl_freetype_glyph_t * glyphs,
                                        int num)
{
    sdl_freetype_text_fixed_extents_t metrics;
    sdl_freetype_bbox_t bbox;
    int i, firsttime = 1;
    int min_x = 0, max_x = 0;
    int min_y = 0, max_y = 0;
    sdl_freetype_fixed_t x_advance = 0, y_advance = 0;

    sdl_freetype_font_glyph_freeze (font);

    for (i = 0; i < num; i++) {
        int                     left, top;
        int                     right, bottom;
        int                     x, y;

        if (sdl_freetype_font_glyph_metrics_and_bbox (font, glyphs[i].index,
                                                      &metrics, &bbox) < 0)
            continue;


        /* glyph images are snapped to pixel locations */
        x = sdl_freetype_fixed_round (glyphs[i].x);
        y = sdl_freetype_fixed_round (glyphs[i].y);

        left   = x + sdl_freetype_fixed_floor (bbox.x1);
        top    = y + sdl_freetype_fixed_floor (bbox.y1);
        right  = x + sdl_freetype_fixed_ceil (bbox.x2);
        bottom = y + sdl_freetype_fixed_ceil (bbox.y2);

        x_advance += metrics.x_advance;
        y_advance += metrics.y_advance;

        if (firsttime) {
            min_x = left;
            max_x = right;
            min_y = top;
            max_y = bottom;
            firsttime = 0;
        } else {
            if (left < min_x)
                min_x = left;
            if (right > max_x)
                max_x = right;
            if (top < min_y)
                min_y = top;
            if (bottom > max_y)
                max_y = bottom;
        }
    }

    extents->x_bearing = sdl_freetype_fixed_from_integer(min_x);
    extents->y_bearing = sdl_freetype_fixed_from_integer(min_y);
    extents->width = sdl_freetype_fixed_from_integer(max_x - min_x);
    extents->height = sdl_freetype_fixed_from_integer(max_y - min_y);
    extents->x_advance = x_advance;
    extents->y_advance = y_advance;

    sdl_freetype_font_glyph_thaw (font);

    return SDL_FREETYPE_ERR_OK;
}

sdl_freetype_error_t
sdl_freetype_font_glyphs_extents (sdl_freetype_font_t * font,
                                  sdl_freetype_text_extents_t * extents,
                                  const sdl_freetype_glyph_t * glyphs,
                                  int num)
{
    sdl_freetype_text_fixed_extents_t fixed;
    int ret;

    ret = sdl_freetype_font_glyphs_fixed_extents(font, &fixed, glyphs, num);
    if (ret < 0)
        return ret;

    sdl_freetype_text_extents_from_fixed(extents, &fixed);
    return 0;
}

sdl_freetype_error_t
sdl_freetype_font_text_fixed_extents (sdl_freetype_font_t * font,
                                      sdl_freetype_text_fixed_extents_t * extents,
                                      const FT_Int32 * ucs4, int num)
{
    sdl_freetype_glyph_t * glyphs, lglyphs[LOCAL_BUFSZ];
    sdl_freetype_text_fixed_extents_t metrics;
    sdl_freetype_fixed_t x_offset, y_offset, x_advance, y_advance;
    FT_Face face;
    int ret, i;

    face = sdl_freetype_font_lock_face (font);
    if (!face)
        goto errquit0;
    if (num <= LOCAL_BUFSZ)
        glyphs = lglyphs;
    else
        glyphs = malloc (sizeof(sdl_freetype_glyph_t) * num);
    if (!glyphs)
        goto errquit1;

    for (i = 0; i < num; i++)
        glyphs[i].index = FT_Get_Char_Index (face, ucs4[i]);

    sdl_freetype_font_unlock_face (font);

    sdl_freetype_font_glyph_freeze (font);

    x_offset = 0;
    y_offset = 0;
    x_advance = y_advance = 0;
    for (i = 0; i < num; i++) {
        x_offset += x_advance;
        y_offset += y_advance;
        glyphs[i].x = x_offset;
        glyphs[i].y = y_offset;
        if (sdl_freetype_font_glyph_fixed_metrics (font, glyphs[i].index,
                                                   &metrics) < 0) {
            x_advance = y_advance = 0;
        } else {
            x_advance = metrics.x_advance;
            y_advance = metrics.y_advance;
        }
    }

    ret = sdl_freetype_font_glyphs_fixed_extents (font, extents, glyphs, num);
    if (glyphs != lglyphs)
        free (glyphs);

    sdl_freetype_font_glyph_thaw (font);

    return ret;
 errquit1:
    sdl_freetype_font_unlock_face (font);
 errquit0:
    return -1;
}

sdl_freetype_error_t
sdl_freetype_font_text_extents (sdl_freetype_font_t * font,
                                sdl_freetype_text_extents_t * extents,
                                const FT_Int32 * ucs4, int num)
{
    sdl_freetype_text_fixed_extents_t fextents;
    int ret;

    ret = sdl_freetype_font_text_fixed_extents (font, &fextents, ucs4, num);
    if (ret < 0)
        return ret;

    extents->x_bearing = sdl_freetype_fixed_to_double(fextents.x_bearing);
    extents->y_bearing = sdl_freetype_fixed_to_double(fextents.y_bearing);
    extents->x_advance = sdl_freetype_fixed_to_double(fextents.x_advance);
    extents->y_advance = sdl_freetype_fixed_to_double(fextents.y_advance);
    extents->width = sdl_freetype_fixed_to_double(fextents.width);
    extents->height = sdl_freetype_fixed_to_double(fextents.height);

    return 0;
}

sdl_freetype_error_t
sdl_freetype_font_show_text (sdl_freetype_font_t * font,
                             void * dst,
                             unsigned char r,
                             unsigned char g,
                             unsigned char b,
                             unsigned char a,
                             int x, int y,
                             const FT_Int32 * ucs4,
                             int num)
{
    sdl_freetype_glyph_t * glyphs, lglyphs[LOCAL_BUFSZ];
    sdl_freetype_text_fixed_extents_t metrics;
    sdl_freetype_fixed_t x_offset, y_offset, x_advance, y_advance;
    FT_Face face;
    int ret, i;

    if (!a)
        return 0;

    face = sdl_freetype_font_lock_face (font);
    if (!face)
        goto errquit0;
    if (num <= LOCAL_BUFSZ)
        glyphs = lglyphs;
    else
        glyphs = malloc (sizeof(sdl_freetype_glyph_t) * num);
    if (!glyphs)
        goto errquit1;

    for (i = 0; i < num; i++)
        glyphs[i].index = FT_Get_Char_Index (face, ucs4[i]);

    sdl_freetype_font_glyph_freeze (font);

    x_offset = 0;
    y_offset = 0;
    x_advance = y_advance = 0;
    for (i = 0; i < num; i++) {
        x_offset += x_advance;
        y_offset += y_advance;
        glyphs[i].x = x_offset;
        glyphs[i].y = y_offset;
        if (sdl_freetype_font_glyph_fixed_metrics (font, glyphs[i].index,
                                                   &metrics) < 0) {
            x_advance = y_advance = 0;
        } else {
            x_advance = metrics.x_advance;
            y_advance = metrics.y_advance;
        }
    }

    sdl_freetype_font_unlock_face (font);

    ret = sdl_freetype_font_show_glyphs (font, dst, r, g, b, a,
                                         x, y, glyphs, num);
    if (glyphs != lglyphs)
        free (glyphs);

    sdl_freetype_font_glyph_thaw (font);

    return ret;
 errquit1:
    sdl_freetype_font_unlock_face (font);
 errquit0:
    return -1;
}

sdl_freetype_error_t
sdl_freetype_font_show_utf8 (sdl_freetype_font_t * font,
                             void * dst,
                             unsigned char r,
                             unsigned char g,
                             unsigned char b,
                             unsigned char a,
                             int x, int y,
                             const char * utf8, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf8_to_ucs4 (utf8, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_show_text (font, dst,
                                       r, g, b, a,
                                       x, y, ucs4, ulen);

    free (ucs4);

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_utf8_extents (sdl_freetype_font_t * font,
                                sdl_freetype_text_extents_t * extents,
                                const char * utf8, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf8_to_ucs4 (utf8, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_text_extents (font, extents, ucs4, ulen);

    free (ucs4);

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_utf8_fixed_extents (sdl_freetype_font_t * font,
                                      sdl_freetype_text_fixed_extents_t * extents,
                                      const char * utf8, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf8_to_ucs4 (utf8, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_text_fixed_extents (font, extents, ucs4, ulen);

    free (ucs4);

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_show_utf16 (sdl_freetype_font_t * font,
			      void * dst,
			      unsigned char r,
			      unsigned char g,
			      unsigned char b,
			      unsigned char a,
			      int x, int y,
			      const short * utf16, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf16_to_ucs4 (utf16, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_show_text (font, dst,
                                       r, g, b, a,
                                       x, y, ucs4, ulen);

    free (ucs4);

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_utf16_extents (sdl_freetype_font_t * font,
				 sdl_freetype_text_extents_t * extents,
				 const short * utf16, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf16_to_ucs4 (utf16, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_text_extents (font, extents, ucs4, ulen);

    free (ucs4);

    return ret;
}

sdl_freetype_error_t
sdl_freetype_font_utf16_fixed_extents (sdl_freetype_font_t * font,
				       sdl_freetype_text_fixed_extents_t * extents,
				       const short * utf16, int len)
{
    FT_Int32 * ucs4;
    int ulen, ret;

    ulen = sdl_freetype_utf16_to_ucs4 (utf16, len, &ucs4);
    if (ulen < 0)
        return -1;

    ret = sdl_freetype_font_text_fixed_extents (font, extents, ucs4, ulen);

    free (ucs4);

    return ret;
}
